# TODO [DG] allow user to set default resource types for user, project, or org,
#  including coarse selections like 'local' or 'aws'
# TODO allow registration of user-created resource implementations

from .blob import Blob
from .hardware import Cluster, cluster, SkyCluster
from .packages.package import Package
from .resource import Resource
from .secrets import Secrets
from .send import Send
from .tables.table import Table



import copy
import json
import logging
from pathlib import Path
from typing import Any, Dict, Optional

import requests
import yaml

from runhouse.logger import LOGGING_CONFIG

from runhouse.rns.api_utils.utils import read_response_data, to_bool

# Configure the logger once
logging.config.dictConfig(LOGGING_CONFIG)

logger = logging.getLogger(__name__)


class Defaults:
    """Class to handle defaults for Runhouse. Defaults are stored in a json file in the user's home directory."""

    USER_ENDPOINT = "user/"
    GROUP_ENDPOINT = "group/"
    CONFIG_PATH = Path("~/.rh/config.yaml").expanduser()
    # TODO [DG] default sub-dicts for various resources (e.g. defaults.get('cluster').get('resource_type'))
    BASE_DEFAULTS = {
        "default_folder": "~",
        "default_provider": "cheapest",
        "default_autostop": -1,
        "use_spot": False,
        "use_local_configs": True,
        "use_rns": False,
        "api_server_url": "https://api.run.house",
    }

    def __init__(self):
        self._defaults_cache = {}

    @property
    def defaults_cache(self):
        if not self._defaults_cache:
            self._defaults_cache = self.load_defaults_from_file()
        return self._defaults_cache

    @defaults_cache.setter
    def defaults_cache(self, value: Dict):
        self._defaults_cache = value

    def load_defaults_from_file(self, config_path: Optional[str] = None) -> Dict:
        config_path = config_path or self.CONFIG_PATH
        config = {}
        if Path(config_path).exists():
            with open(config_path, "r") as stream:
                config = yaml.safe_load(stream)
            logging.info(f"Loaded Runhouse config from {config_path}")

        return config or {}

    @property
    def request_headers(self):
        return {"Authorization": f"Bearer {self.get('token')}"}

    def upload_defaults(
        self,
        defaults: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        entity: Optional[str] = "user",
    ):
        """Upload defaults into rns. If defaults is None, upload the defaults from the local config file,
        `~/.rh/config.yaml."""
        if not defaults:
            defaults = self.defaults_cache
            self.save_defaults()
        to_upload = copy.deepcopy(defaults)
        # We don't need to save these
        to_upload.pop("token", None)
        to_upload.pop("username", None)

        endpoint = (
            self.USER_ENDPOINT
            if entity == "user"
            else self.GROUP_ENDPOINT + f"/{entity}"
        )
        resp = requests.put(
            f'{self.get("api_server_url")}/{endpoint}/config',
            data=json.dumps(to_upload),
            headers=headers or self.request_headers,
        )
        if resp.status_code != 200:
            raise Exception(
                f"Failed to update defaults for {entity}, received status code {resp.status_code}"
            )
        logger.info(f"Uploaded defaults for {entity} to rns.")

    def download_defaults(
        self, headers: Optional[Dict] = None, entity: Optional[str] = "user"
    ) -> Dict:
        """Get defaults for user or group."""
        endpoint = (
            self.USER_ENDPOINT
            if entity == "user"
            else self.GROUP_ENDPOINT + f"/{entity}"
        )
        headers = headers or self.request_headers
        resp = requests.get(f'{self.get("api_server_url")}/{endpoint}', headers=headers)
        if resp.status_code != 200:
            raise Exception(
                f"Failed to download defaults for {entity}, received status code {resp.status_code}"
            )
        resp_data: dict = read_response_data(resp)
        raw_defaults = resp_data.get("config", {})
        raw_defaults["username"] = resp_data.get("username")
        raw_defaults["token"] = headers.get("Authorization")[7:]
        raw_defaults["default_folder"] = (
            "/" + raw_defaults["username"]
            if raw_defaults.get("default_folder") in ["~", None]
            else raw_defaults.get("default_folder")
        )
        formatted = {k: to_bool(v) for k, v in raw_defaults.items()}

        return formatted

    def save_defaults(
        self, defaults: Optional[Dict] = None, config_path: Optional[str] = None
    ):
        config_path = Path(config_path or self.CONFIG_PATH)
        defaults = defaults or self.defaults_cache

        if not config_path.exists():
            config_path.parent.mkdir(parents=True, exist_ok=True)

        with config_path.open("w") as stream:
            yaml.dump(defaults, stream)

    def download_and_save_defaults(
        self,
        headers: Optional[Dict] = None,
        entity: Optional[str] = "user",
        config_path: Optional[str] = None,
    ):
        """Download defaults from rns and save them to the local config file."""
        defaults = self.download_defaults(headers=headers, entity=entity)
        # Note: downloaded defaults take priority over local defaults
        self.set_many(defaults, config_path=config_path)

    def set(self, key: str, value: Any, config_path: Optional[str] = None):
        self.defaults_cache[key] = value
        self.save_defaults(config_path=config_path)

    def set_many(self, key_value_pairs: Dict, config_path: Optional[str] = None):
        self.defaults_cache.update(key_value_pairs)
        self.save_defaults(config_path=config_path)

    # TODO [DG] allow hierarchical defaults from folders and groups
    def get(self, key: str, alt: Any = None) -> Any:
        res = self.defaults_cache.get(key, alt)
        if not res and key in self.BASE_DEFAULTS:
            res = self.BASE_DEFAULTS[key]
        return res



import copy
import inspect
import json
import logging
import os
import re
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Union

import grpc
import requests
import sshtunnel

from runhouse import rh_config
from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.api_utils.utils import is_jsonable, read_response_data
from runhouse.rns.hardware import Cluster
from runhouse.rns.packages import git_package, Package

from runhouse.rns.resource import Resource
from runhouse.rns.run_module_utils import call_fn_by_type, get_fn_by_name

logger = logging.getLogger(__name__)


class Send(Resource):
    RESOURCE_TYPE = "send"
    DEFAULT_ACCESS = "write"

    def __init__(
        self,
        fn_pointers: Tuple[str, str, str],
        hardware: Optional[Cluster] = None,
        name: [Optional[str]] = None,
        reqs: Optional[List[str]] = None,
        setup_cmds: Optional[List[str]] = None,
        dryrun: bool = False,
        access: Optional[str] = None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Create, load, or update a Send ("Serverless endpoint"). A Send is comprised of the
        entrypoint, hardware, and dependencies to run the service.

        Args:
            fn (): A python callable or entrypoint string (module:function) within the package which the user
                can call to run remotely as a microservice. The Send object is callable, taking the same inputs
                 and producing the same outputs as fn. For example, if we create
                 `my_send = Send(fn=lambda x: x+1, hardware=my_hw)`, we can call it with `my_send(4)`, and
                 the fn will run remotely on `my_hw`.
            name (): A URI to persist this Send's metadata to Runhouse's Resource Naming System (RNS), which allows
                it to be reloaded and used later or in other environments. If name is left empty, the Send is
                "anonymous" and will only exist within this Python context. # TODO more about user namespaces.
            hardware ():
            reqs ():
            cluster ():
        """
        self.fn_pointers = fn_pointers
        self.hardware = hardware
        self.reqs = reqs
        self.setup_cmds = setup_cmds or []
        self.access = access or self.DEFAULT_ACCESS
        self.dryrun = dryrun
        super().__init__(name=name, dryrun=dryrun)

        if not self.dryrun and self.hardware and self.access in ["write", "read"]:
            self.to(self.hardware, reqs=self.reqs, setup_cmds=setup_cmds)

    # ----------------- Constructor helper methods -----------------

    @staticmethod
    def from_config(config, dryrun=True):
        """Create a Send object from a config dictionary.

        Args:
            config (dict): Dictionary of config values.

        Returns:
            Send: Send object created from config values.
        """
        config["reqs"] = [
            Package.from_config(package, dryrun=True)
            if isinstance(package, dict)
            else package
            for package in config["reqs"]
        ]
        # TODO validate which fields need to be present in the config

        if isinstance(config["hardware"], dict):
            config["hardware"] = Cluster.from_config(config["hardware"], dryrun=dryrun)

        if "fn_pointers" not in config:
            raise ValueError(
                "No fn_pointers provided in config. Please provide a path "
                "to a python file, module, and function name."
            )

        return Send(**config, dryrun=dryrun)

    def to(self, hardware, reqs=None, setup_cmds=None):
        hw_backup = self.hardware
        self.hardware = None
        new_send = copy.deepcopy(self)
        self.hardware = hw_backup
        new_send.hardware = hardware
        new_send.reqs = reqs if reqs else self.reqs
        new_send.setup_cmds = (
            setup_cmds if setup_cmds else self.setup_cmds
        )  # Run inside reup_cluster
        # TODO [DG] figure out how to run setup_cmds on BYO Cluster

        logging.info("Setting up Send on cluster.")
        if not new_send.hardware.address:
            # For SkyCluster, this initial check doesn't trigger a sky.status, which is slow.
            # If cluster simply doesn't have an address we likely need to up it.
            if not hasattr(new_send.hardware, "up"):
                raise ValueError(
                    "Cluster must have an address (i.e. be up) or have a reup_cluster method "
                    "(e.g. SkyCluster)."
                )
            if not new_send.hardware.is_up():
                # If this is a SkyCluster, before we up the cluster, run a sky.check to see if the cluster
                # is already up but doesn't have an address assigned yet.
                new_send.reup_cluster()
        try:
            new_send.hardware.install_packages(new_send.reqs)
        except (grpc.RpcError, sshtunnel.BaseSSHTunnelForwarderError):
            # It's possible that the cluster went down while we were trying to install packages.
            if not new_send.hardware.is_up():
                new_send.reup_cluster()
            else:
                new_send.hardware.restart_grpc_server(resync_rh=False)
            new_send.hardware.install_packages(new_send.reqs)
        logging.info("Send setup complete.")

        return new_send

    def reup_cluster(self):
        logger.info(f"Upping the cluster {self.hardware.name}")
        self.hardware.up()
        # TODO [DG] this only happens when the cluster comes up, not when a new send is added to the cluster
        self.hardware.run(self.setup_cmds)

    def run_setup(self, cmds, force=False):
        to_run = []
        for cmd in cmds:
            if force or cmd not in self.setup_cmds:
                to_run.append(cmd)
        if to_run:
            self.setup_cmds.extend(to_run)
            self.hardware.run(to_run)

    @staticmethod
    def extract_fn_paths(raw_fn, reqs):
        """Get the path to the module, module name, and function name to be able to import it on the server"""
        if not isinstance(raw_fn, Callable):
            raise TypeError(
                f"Invalid fn for Send, expected Callable but received {type(raw_fn)}"
            )
        # Background on all these dunders: https://docs.python.org/3/reference/import.html
        module = inspect.getmodule(raw_fn)

        # Need to resolve in case just filename is given
        module_path = (
            str(Path(inspect.getfile(module)).resolve())
            if hasattr(module, "__file__")
            else None
        )

        if not module_path or raw_fn.__name__ == "<lambda>":
            # The only time __file__ wouldn't be present is if the function is defined in an interactive
            # interpreter or a notebook. We can't import on the server in that case, so we need to cloudpickle
            # the fn to send it over. The __call__ function will serialize the function if we return it this way.
            # This is a short-term hack.
            # return None, "notebook", raw_fn.__name__
            root_path = os.getcwd()
            module_name = "notebook"
            fn_name = raw_fn.__name__
        else:
            root_path = os.path.dirname(module_path)
            module_name = inspect.getmodulename(module_path)
            # TODO __qualname__ doesn't work when fn is aliased funnily, like torch.sum
            fn_name = getattr(raw_fn, "__qualname__", raw_fn.__name__)

            # Adapted from https://github.com/modal-labs/modal-client/blob/main/modal/_function_utils.py#L94
            if getattr(module, "__package__", None):
                module_path = os.path.abspath(module.__file__)
                package_paths = [
                    os.path.abspath(p) for p in __import__(module.__package__).__path__
                ]
                base_dirs = [
                    base_dir
                    for base_dir in package_paths
                    if os.path.commonpath((base_dir, module_path)) == base_dir
                ]

                if len(base_dirs) != 1:
                    logger.info(f"Module files: {module_path}")
                    logger.info(f"Package paths: {package_paths}")
                    logger.info(f"Base dirs: {base_dirs}")
                    raise Exception("Wasn't able to find the package directory!")
                root_path = os.path.dirname(base_dirs[0])
                module_name = module.__spec__.name

        remote_import_path = None
        for req in reqs:
            local_path = None
            if not isinstance(req, str) and req.is_local():
                local_path = Path(req.local_path)
            elif isinstance(req, str):
                if req.split(":")[0] in ["local", "reqs", "pip"]:
                    req = req.split(":")[1]

                if Path(req).expanduser().resolve().exists():
                    # Relative paths are relative to the working directory in Folders/Packages!
                    local_path = (
                        Path(req).expanduser()
                        if Path(req).expanduser().is_absolute()
                        else Path(rh_config.rns_client.locate_working_dir()) / req
                    )

            if local_path:
                try:
                    # Module path relative to package
                    remote_import_path = str(
                        local_path.name / Path(root_path).relative_to(local_path)
                    )
                except ValueError:  # Not a subdirectory
                    pass
        return remote_import_path, module_name, fn_name

    # Found in python decorator logic, maybe use
    # func_name = getattr(f, '__qualname__', f.__name__)
    # module_name = getattr(f, '__module__', '')
    # if module_name:
    #     full_name = f'{module_name}.{func_name}'
    # else:
    #     full_name = func_name

    # ----------------- Send call methods -----------------

    def __call__(self, *args, stream_logs=False, **kwargs):
        fn_type = "call"
        if self.access in [ResourceAccess.write, ResourceAccess.read]:
            if (
                not self.hardware
                or self.hardware.name == rh_config.obj_store.cluster_name
            ):
                [relative_path, module_name, fn_name] = self.fn_pointers
                fn = get_fn_by_name(module_name=module_name, fn_name=fn_name)
                return call_fn_by_type(
                    fn, fn_type, fn_name, relative_path, args, kwargs
                )
            elif stream_logs:
                run_key = self.remote(*args, **kwargs)
                return self.hardware.get(run_key, stream_logs=True)
            else:
                return self._call_fn_with_ssh_access(
                    fn_type=fn_type, args=args, kwargs=kwargs
                )
        else:
            # run the function via http url - user only needs Proxy access
            if self.access != ResourceAccess.proxy:
                raise RuntimeError("Running http url requires proxy access")
            if not rh_config.rns_client.token:
                raise ValueError(
                    "Token must be saved in the local .rh config in order to use an http url"
                )
            http_url = self.http_url()
            logger.info(f"Running {self.name} via http url: {http_url}")
            resp = requests.post(
                http_url,
                data=json.dumps({"args": args, "kwargs": kwargs}),
                headers=rh_config.rns_client.request_headers,
            )
            if resp.status_code != 200:
                raise Exception(
                    f"Failed to run Send endpoint: {json.loads(resp.content)}"
                )

            res = read_response_data(resp)
            return res

    def repeat(self, num_repeats, *args, **kwargs):
        """Repeat the Send call multiple times.

        Args:
            num_repeats (int): Number of times to repeat the Send call.
            *args: Positional arguments to pass to the Send
            **kwargs: Keyword arguments to pass to the Send
        """
        if self.access in [ResourceAccess.write, ResourceAccess.read]:
            return self._call_fn_with_ssh_access(
                fn_type="repeat", args=[num_repeats, args], kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Send.repeat only works with Write or Read access, not Proxy access"
            )

    def map(self, arg_list, **kwargs):
        """Map a function over a list of arguments."""
        if self.access in [ResourceAccess.write, ResourceAccess.read]:
            return self._call_fn_with_ssh_access(
                fn_type="map", args=arg_list, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Send.map only works with Write or Read access, not Proxy access"
            )

    def starmap(self, args_lists, **kwargs):
        """Like Send.map() except that the elements of the iterable are expected to be iterables
        that are unpacked as arguments. An iterable of [(1,2), (3, 4)] results in [func(1,2), func(3,4)]."""
        if self.access in [ResourceAccess.write, ResourceAccess.read]:
            return self._call_fn_with_ssh_access(
                fn_type="starmap", args=args_lists, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Send.starmap only works with Write or Read access, not Proxy access"
            )

    def enqueue(self, *args, **kwargs):
        """Enqueue a Send call to be run later.

        Args:
            *args: Positional arguments to pass to the Send
            **kwargs: Keyword arguments to pass to the Send
        """
        if self.access in [ResourceAccess.write, ResourceAccess.read]:
            return self._call_fn_with_ssh_access(
                fn_type="queue", args=args, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Send.enqueue only works with Write or Read access, not Proxy access"
            )

    def remote(self, *args, **kwargs):
        """Map a function over a list of arguments."""
        # TODO [DG] pin the obj_ref and return a string (printed to log) so result can be retrieved later and we
        # don't need to init ray here. Also, allow user to pass the string as a param to remote().
        # TODO [DG] add rpc for listing gettaable strings, plus metadata (e.g. when it was created)
        # We need to ray init here so the returned Ray object ref doesn't throw an error it's deserialized
        # import ray
        # ray.init(ignore_reinit_error=True)
        if self.access in [ResourceAccess.write, ResourceAccess.read]:
            run_key = self._call_fn_with_ssh_access(
                fn_type="remote", args=args, kwargs=kwargs
            )
            cluster_name = (
                f'rh.cluster(name="{self.hardware.rns_address}")'
                if self.hardware.name
                else "<my_cluster>"
            )
            logger.info(
                f"Submitted remote call to cluster. Result or logs can be retrieved"
                f'\n with run_key "{run_key}", e.g. '
                f'\n`{cluster_name}.get("{run_key}", stream_logs=True)` in python '
                f"\n or cancelled with "
                f'\n`{cluster_name}.cancel("{run_key}")` in python or '
                f'\n`runhouse cancel "{cluster_name}" {run_key}` from the command line.'
            )
            return run_key
        else:
            raise NotImplementedError(
                "Send.remote only works with Write or Read access, not Proxy access"
            )

    def get(self, obj_ref):
        """Get the result of a Send call that was submitted as async using `remote`.

        Args:
            obj_ref: A single or list of Ray.ObjectRef objects returned by a Send.remote() call. The ObjectRefs
                must be from the cluster that this Send is running on.
        """
        # TODO [DG] replace with self.hardware.get()?
        if self.access in [ResourceAccess.write, ResourceAccess.read]:
            arg_list = obj_ref if isinstance(obj_ref, list) else [obj_ref]
            return self._call_fn_with_ssh_access(
                fn_type="get", args=arg_list, kwargs={}
            )
        else:
            raise NotImplementedError(
                "Send.get only works with Write or Read access, not Proxy access"
            )

    def _call_fn_with_ssh_access(self, fn_type, args, kwargs):
        # https://docs.ray.io/en/latest/ray-core/tasks/patterns/map-reduce.html
        # return ray.get([map.remote(i, map_func) for i in replicas])
        # TODO allow specifying resources per worker for map
        # TODO [DG] check whether we're on the cluster and if so, just call the function directly via the
        # helper function currently in UnaryServer
        name = self.name or "anonymous send"
        if self.fn_pointers is None:
            raise RuntimeError(f"No fn pointers saved for {name}")

        [relative_path, module_name, fn_name] = self.fn_pointers
        name = self.name or fn_name or "anonymous send"
        logger.info(f"Running {name} via gRPC")
        res = self.hardware.run_module(
            relative_path, module_name, fn_name, fn_type, args, kwargs
        )
        return res

    # TODO [DG] test this properly
    # def debug(self, redirect_logging=False, timeout=10000, *args, **kwargs):
    #     """Run the Send in debug mode. This will run the Send through a tunnel interpreter, which
    #     allows the use of breakpoints and other debugging tools, like rh.ipython().
    #     FYI, alternative ideas from Ray: https://github.com/ray-project/ray/issues/17197
    #     FYI, alternative Modal folks shared: https://github.com/modal-labs/modal-client/pull/32
    #     """
    #     from paramiko import AutoAddPolicy
    #
    #     # Importing this here because they're heavy
    #     from plumbum.machines.paramiko_machine import ParamikoMachine
    #     from rpyc.utils.classic import redirected_stdio
    #     from rpyc.utils.zerodeploy import DeployedServer
    #
    #     creds = self.hardware.ssh_creds()
    #     ssh_client = ParamikoMachine(
    #         self.hardware.address,
    #         user=creds["ssh_user"],
    #         keyfile=str(Path(creds["ssh_private_key"]).expanduser()),
    #         missing_host_policy=AutoAddPolicy(),
    #     )
    #     server = DeployedServer(
    #         ssh_client, server_class="rpyc.utils.server.ForkingServer"
    #     )
    #     conn = server.classic_connect()
    #
    #     if redirect_logging:
    #         rlogger = conn.modules.logging.getLogger()
    #         rlogger.parent = logging.getLogger()
    #
    #     conn._config[
    #         "sync_request_timeout"
    #     ] = timeout  # seconds. May need to be longer for real debugging.
    #     conn._config["allow_public_attrs"] = True
    #     conn._config["allow_pickle"] = True
    #     # This assumes the code is already synced over to the remote container
    #     remote_fn = getattr(conn.modules[self.fn.__module__], self.fn.__name__)
    #
    #     with redirected_stdio(conn):
    #         res = remote_fn(*args, **kwargs)
    #     conn.close()
    #     return res

    @property
    def config_for_rns(self):
        # TODO save Package resource, because fn_pointers are meaningless without the package.

        config = super().config_for_rns

        config.update(
            {
                "hardware": self._resource_string_for_subconfig(self.hardware.save()),
                "reqs": [
                    self._resource_string_for_subconfig(package)
                    for package in self.reqs
                ],
                "setup_cmds": self.setup_cmds,
                "fn_pointers": self.fn_pointers,
            }
        )
        return config

    # TODO maybe reuse these if we starting putting each send in its own container
    # @staticmethod
    # def run_ssh_cmd_in_cluster(ssh_key, ssh_user, address, cmd, port_fwd=None):
    #     subprocess.run("ssh -tt -o IdentitiesOnly=yes -i "
    #                    f"{ssh_key} {port_fwd or ''}"
    #                    f"{ssh_user}@{address} docker exec -it ray_container /bin/bash -c {cmd}".split(' '))

    def ssh(self):
        if self.hardware is None:
            raise RuntimeError("Hardware must be specified and up to ssh into a Send")
        self.hardware.ssh()

    def send_secrets(self, reload=False):
        self.hardware.send_secrets(reload=reload)

    def http_url(self, curl_command=False, *args, **kwargs) -> str:
        """Return the endpoint needed to run the Send on the remote cluster, or provide the curl command if requested"""
        resource_uri = rh_config.rns_client.resource_uri(name=self.name)
        uri = f"proxy/{resource_uri}"
        if curl_command:
            # NOTE: curl command should include args and kwargs - this will help us generate better API docs
            if not is_jsonable(args) or not is_jsonable(kwargs):
                raise Exception(
                    "Invalid Send func params provided, must be able to convert args and kwargs to json"
                )

            return (
                "curl -X 'POST' '{api_server_url}/proxy{resource_uri}/endpoint' "
                "-H 'accept: application/json' "
                "-H 'Authorization: Bearer {auth_token}' "
                "-H 'Content-Type: application/json' "
                "-d '{data}'".format(
                    api_server_url=rh_config.rns_client.api_server_url,
                    resource_uri=uri,
                    auth_token=rh_config.rns_client.token,
                    data=json.dumps({"args": args, "kwargs": kwargs}),
                )
            )

        # HTTP URL needed to run the Send remotely
        http_url = f"{rh_config.rns_client.api_server_url}/{uri}/endpoint"
        return http_url

    def notebook(self, persist=False, sync_package_on_close=None, port_forward=8888):
        # Roughly trying to follow:
        # https://towardsdatascience.com/using-jupyter-notebook-running-on-a-remote-docker-container-via-ssh-ea2c3ebb9055
        # https://docs.ray.io/en/latest/ray-core/using-ray-with-jupyter.html
        if self.hardware is None:
            raise RuntimeError("Cannot SSH, running locally")

        tunnel, port_fwd = self.hardware.ssh_tunnel(
            local_port=port_forward, num_ports_to_try=10
        )
        try:
            install_cmd = "pip install jupyterlab"
            jupyter_cmd = f"jupyter lab --port {port_fwd} --no-browser"
            # port_fwd = '-L localhost:8888:localhost:8888 '  # TOOD may need when we add docker support
            with self.hardware.pause_autostop():
                self.hardware.run(commands=[install_cmd, jupyter_cmd], stream_logs=True)

        finally:
            if sync_package_on_close:
                if sync_package_on_close == "./":
                    sync_package_on_close = rh_config.rns_client.locate_working_dir()
                pkg = Package.from_string("local:" + sync_package_on_close)
                self.hardware.rsync(
                    source=f"~/{pkg.name}", dest=pkg.local_path, up=False
                )
            if not persist:
                tunnel.stop(force=True)
                kill_jupyter_cmd = f"jupyter notebook stop {port_fwd}"
                self.hardware.run(commands=[kill_jupyter_cmd])

    def keep_warm(
        self,
        autostop_mins=None,
        # TODO regions: List[str] = None,
        # TODO min_replicas: List[int] = None,
        # TODO max_replicas: List[int] = None
    ):
        if autostop_mins is None:
            logger.info(f"Keeping {self.name} indefinitely warm")
            # keep indefinitely warm if user doesn't specify
            autostop_mins = -1
        self.hardware.keep_warm(autostop_mins=autostop_mins)

    @staticmethod
    def _handle_nb_fn(fn, fn_pointers, serialize_notebook_fn, name):
        """Handle the case where the user passes in a notebook function"""
        if serialize_notebook_fn:
            # This will all be cloudpickled by the gRPC client and unpickled by the gRPC server
            # Note that this means the function cannot be saved, and it's better that way because
            # pickling functions is not meant for long term storage. Case in point, this method will be
            # sensitive to differences in minor Python versions between the serializing and deserializing envs.
            return "", "notebook", fn
        else:
            # TODO put this in the current folder instead?
            module_path = Path.cwd() / (f"{name}_fn.py" if name else "send_fn.py")
            logging.info(
                f"Writing out send function to {str(module_path)} as "
                f"functions serialized in notebooks are brittle. Please make "
                f"sure the function does not rely on any local variables, "
                f"including imports (which should be moved inside the function body)."
            )
            if not name:
                logging.warning(
                    "You should name Sends that are created in notebooks to avoid naming collisions "
                    "between the modules that are created to hold their functions "
                    '(i.e. "send_fn.py" errors.'
                )
            source = inspect.getsource(fn).strip()
            with module_path.open("w") as f:
                f.write(source)
            return fn_pointers[0], module_path.stem, fn_pointers[2]
            # from importlib.util import spec_from_file_location, module_from_spec
            # spec = spec_from_file_location(config['name'], str(module_path))
            # module = module_from_spec(spec)
            # spec.loader.exec_module(module)
            # new_fn = getattr(module, fn_pointers[2])
            # fn_pointers = Send.extract_fn_paths(raw_fn=new_fn, reqs=config['reqs'])


def send(
    fn: Optional[Union[str, Callable]] = None,
    name: [Optional[str]] = None,
    hardware: Optional[Union[str, Cluster]] = None,
    reqs: Optional[List[str]] = None,
    setup_cmds: Optional[List[str]] = None,
    # TODO image: Optional[str] = None,
    dryrun: bool = False,
    load_secrets: bool = False,
    serialize_notebook_fn: bool = False,
):
    """Factory constructor to construct the Send for various provider types.

    fn: The function which will execute on the remote cluster when this send is called.
    name: Name of the Send to create or retrieve, either from a local config or from the RNS.
    hardware: Hardware to use for the Send, either a string name of a Cluster object, or a Cluster object.
    package: Package to send to the remote cluster, either a string name of a Package, package url,
        or a Package object.
    reqs: List of requirements to install on the remote cluster, or path to a requirements.txt file. If a list
        of pypi packages is provided, including 'requirements.txt' in the list will install the requirements
        in `package`. By default, if reqs is left as None, we'll set it to ['requirements.txt'], which installs
        just the requirements of package. If an empty list is provided, no requirements will be installed.
    image (TODO): Docker image id to use on the remote cluster, or path to Dockerfile.
    dryrun: Whether to create the Send if it doesn't exist, or load the Send object as a dryrun.
    """

    config = rh_config.rns_client.load_config(name)
    config["name"] = name or config.get("rns_address", None) or config.get("name")
    config["reqs"] = reqs if reqs is not None else config.get("reqs", [])

    processed_reqs = []
    for req in config["reqs"]:
        # TODO [DG] the following is wrong. RNS address doesn't have to start with '/'. However if we check if each
        #  string exists in RNS this will be incredibly slow, so leave it for now.
        if isinstance(req, str) and req[0] == "/" and rh_config.rns_client.exists(req):
            # If req is an rns address
            req = rh_config.rns_client.load_config(req)
        processed_reqs.append(req)
    config["reqs"] = processed_reqs

    if callable(fn):
        if not [req for req in config["reqs"] if "./" in req]:
            config["reqs"].append("./")
        fn_pointers = Send.extract_fn_paths(raw_fn=fn, reqs=config["reqs"])
        if fn_pointers[1] == "notebook":
            fn_pointers = Send._handle_nb_fn(
                fn,
                fn_pointers=fn_pointers,
                serialize_notebook_fn=serialize_notebook_fn,
                name=config["name"],
            )
        config["fn_pointers"] = fn_pointers
    elif isinstance(fn, str):
        # Url must match a regex of the form
        # 'https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name'
        # Use a regex to extract username, repo_name, branch_name, path/to/file.py, and func_name
        pattern = (
            r"https://github\.com/(?P<username>[^/]+)/(?P<repo_name>[^/]+)/blob/"
            r"(?P<branch_name>[^/]+)/(?P<path>[^:]+):(?P<func_name>.+)"
        )
        match = re.match(pattern, fn)

        if match:
            username = match.group("username")
            repo_name = match.group("repo_name")
            branch_name = match.group("branch_name")
            path = match.group("path")
            func_name = match.group("func_name")
        else:
            raise ValueError(
                "fn must be a callable or string of the form "
                '"https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name"'
            )
        module_name = Path(path).stem
        relative_path = str(repo_name / Path(path).parent)
        config["fn_pointers"] = (relative_path, module_name, func_name)
        # TODO [DG] check if the user already added this in their reqs
        repo_package = git_package(
            git_url=f"https://github.com/{username}/{repo_name}.git",
            revision=branch_name,
        )
        config["reqs"].insert(0, repo_package)
        # repo_package = Package(url=f'/',
        #                        fs='github',
        #                        data_config={'org': username, 'repo': repo_name, 'sha': branch_name,
        #                                     'filecache': {'cache_storage': repo_name}},
        #                        install_method='local')
        # config['reqs'] = [repo_package] + config['reqs']

    config["hardware"] = hardware or config.get("hardware")
    if isinstance(config["hardware"], str):
        hw_dict = rh_config.rns_client.load_config(config["hardware"])
        if not hw_dict:
            raise RuntimeError(
                f'Hardware {rh_config.rns_client.resolve_rns_path(config["hardware"])} '
                f"not found locally or in RNS."
            )
        config["hardware"] = hw_dict

    config["setup_cmds"] = (
        setup_cmds if setup_cmds is not None else config.get("setup_cmds")
    )

    config["access_level"] = config.get("access_level", Send.DEFAULT_ACCESS)

    new_send = Send.from_config(config, dryrun=dryrun)

    if load_secrets and not dryrun:
        new_send.send_secrets()

    return new_send



import unittest

from runhouse.rns.hardware import cluster, SkyCluster


def test_cluster_config():
    rh_cpu = cluster(name="^rh-cpu", dryrun=False)
    if not rh_cpu.is_up():
        rh_cpu.up()
    config = rh_cpu.config_for_rns
    cluster2 = SkyCluster.from_config(config)
    assert cluster2.address == rh_cpu.address


def test_cluster_sharing():
    # TODO [DG] finish
    pass


def test_install():
    c = cluster(name="^rh-cpu")
    c.install_packages(
        [
            "./",
            "torch==1.12.1",
            # 'conda:jupyterlab',  # TODO [DG] make this actually work
            # 'gh:pytorch/vision'  # TODO [DG] make this actually work
        ]
    )


def test_basic_run():
    # Create temp file where fn's will be stored
    test_cmd = "echo hi"
    hw = cluster(name="^rh-cpu")
    hw.up_if_not()
    res = hw.run(commands=[test_cmd])
    assert "hi" in res[0][1]


def test_restart_grpc():
    hw = cluster(name="^rh-cpu")
    hw.up_if_not()
    codes = hw.restart_grpc_server(resync_rh=False)
    assert codes


if __name__ == "__main__":
    unittest.main()



import unittest

import runhouse as rh


def test_get_all_secrets():
    secrets = rh.Secrets.download_into_env(save_locally=False)
    providers = rh.Secrets.enabled_providers(as_str=True)
    # TODO this check is dependent on local secrets config, not a catch all
    assert set(providers) == {"aws", "gcp", "lambda", "ssh", "huggingface"}
    assert secrets


def test_get_custom_provider_secrets():
    provider = "snowflake"
    provider_secrets = rh.Secrets.get(provider=provider)
    assert provider_secrets


def test_get_provider_secrets():
    provider = "aws"
    provider_secrets = rh.Secrets.get(provider=provider)
    assert provider_secrets


def test_upload_group_provider_secrets():
    provider = "aws"
    provider_secrets = rh.Secrets.put(provider=provider, group="ds_preproc2")
    assert provider_secrets


def test_upload_user_provider_custom_secrets():
    provider = "snowflake"
    rh.Secrets.put(provider=provider, secret={"secret_key": "abcdefg"})
    assert True


def test_upload_user_provider_enabled_secrets():
    provider = "aws"
    rh.Secrets.put(provider=provider)
    assert True


def test_upload_all_provider_secrets():
    rh.Secrets.extract_and_upload(interactive=False)
    assert True


def test_delete_provider_secrets():
    rh.Secrets.delete(providers=["huggingface"])
    assert True


@unittest.skip("This test overrides local rh token if done incorrectly")
# Running this unit test will override local rh config token to "..."
# Was meant to be run by manually inputting token, but need a better way to test
def test_login():
    # TODO [DG] create a mock account and test this properly in CI
    token = "..."

    rh.login(
        token=token,
        download_config=False,
        download_secrets=False,
        upload_secrets=False,
        upload_config=False,
    )

    rh.login(
        token=token,
        download_config=False,
        download_secrets=False,
        upload_secrets=True,
        upload_config=True,
    )

    rh.login(
        token=token,
        download_config=True,
        download_secrets=True,
        upload_secrets=False,
        upload_config=False,
    )

    rh.login(
        token=token,
        download_config=True,
        download_secrets=True,
        upload_secrets=True,
        upload_config=True,
    )

    assert rh.rns_client.default_folder == "/..."


if __name__ == "__main__":
    unittest.main()



import sys
from typing import List

from runhouse.rh_config import rns_client


def resolve_rns_path(path: str):
    return rns_client.resolve_rns_path(path)


def exists(name, resource_type: str = None):
    return rns_client.exists(name, resource_type=resource_type)


def locate(name_or_path, resolve_path: bool = True):
    return rns_client.locate(
        name_or_path,
        resolve_path=resolve_path,
    )


def load(name: str, instantiate: bool = True, dryrun: bool = False):
    config = rns_client.load_config(name=name)
    if not instantiate:
        return config
    from_config_constructor = getattr(
        sys.modules["runhouse.rns"], config["resource_type"].capitalize(), None
    ).from_config
    if not from_config_constructor:
        raise ValueError(
            f"Could not find constructor for type {config['resource_type']}"
        )
    return from_config_constructor(config=config, dryrun=dryrun)


def load_from_path(
    path: str,
    instantiate: bool = True,
):
    pass


def set_save_to(save_to: List[str]):
    rns_client.save_to = save_to


def set_load_from(load_from: List[str]):
    rns_client.load_from = load_from


def save(
    resource,
    name: str = None,
    snapshot: bool = False,
    overwrite: bool = True,
    **snapshot_kwargs,
):  # TODO [DG] was this supposed to be kwargs for the snapshot?
    """Register the resource, saving it to local working_dir config and/or RNS config store. Uses the resource's
    `self.config_for_rns` to generate the dict to save."""

    # TODO handle self.access == 'read' instead of this weird overwrite argument
    snapshot_kwargs = snapshot_kwargs or {}
    resource_to_save = resource.snapshot(**snapshot_kwargs) if snapshot else resource
    if name:
        if "/" in name[1:] or resource_to_save._rns_folder is None:
            (
                resource_to_save._name,
                resource_to_save._rns_folder,
            ) = split_rns_name_and_path(resolve_rns_path(name))
        else:
            resource_to_save._name = name
    rns_client.save_config(resource=resource_to_save, overwrite=overwrite)


def set_folder(path: str, create=False):
    rns_client.set_folder(path=path, create=create)


def unset_folder():
    """Sort of like `cd -`, but with a full stack of the previous folder's set. Resets the
    current_folder to the previous one on the stack, the current_folder right before the
    current one was set."""
    rns_client.unset_folder()


def current_folder():
    return rns_client.current_folder


def split_rns_name_and_path(path: str):
    return rns_client.split_rns_name_and_path(path)


def resources(path: str = None, full_paths=False):
    path = path or current_folder()
    return rns_client.contents(name_or_path=path, full_paths=full_paths)


def ipython():
    import subprocess

    subprocess.Popen("pip install ipython".split(" "))
    # TODO install ipython if not installed
    import IPython

    IPython.embed()


def delete(resource_or_name: str):
    """Delete the resource from the RNS or local config store."""
    rns_client.delete_configs(resource=resource_or_name)


# TODO [DG]
def delete_all(folder: str = None):
    """Delete all resources in the given folder, such that the user has peace of mind that they are not consuming
    any hidden cloud costs."""
    pass


# TODO [DG]
def sync_down():
    pass


# TODO [DG]
def load_all_clusters():
    """Load all clusters in RNS into the local Sky context."""
    pass


# -----------------  Pinning objects to cluster memory  -----------------
# TODO is this a bad idea?

from runhouse import rh_config


def pin_to_memory(key: str, value):
    rh_config.obj_store.put(key, value)


def get_pinned_object(key: str, default=None):
    return rh_config.obj_store.get(key, default=default)


def get(key: str, cluster=None, default=None):
    from runhouse.rns.hardware.skycluster import SkyCluster

    if isinstance(cluster, str):
        if cluster == rh_config.obj_store.cluster_name:
            # We're currently on cluster, so just get the object from local rh_config.obj_store
            return rh_config.obj_store.get(key, default=default)
        else:
            cluster = SkyCluster.from_name(cluster)

    if cluster.name == rh_config.obj_store.cluster_name:
        return rh_config.obj_store.get(key, default=default)
    else:
        return cluster.get(key, default=default)


def remove_pinned_object(key: str):
    rh_config.obj_store.delete(key)


def pop_pinned_object(key: str, default=None):
    return rh_config.obj_store.pop(key, default=default)


def pinned_keys():
    return rh_config.obj_store.keys()


def clear_pinned_memory():
    rh_config.obj_store.clear()



import json
import logging

import pprint
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import requests

from runhouse.rh_config import rns_client
from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.api_utils.utils import read_response_data
from runhouse.rns.top_level_rns_fns import (
    resolve_rns_path,
    save,
    split_rns_name_and_path,
)

logger = logging.getLogger(__name__)


class Resource:
    RESOURCE_TYPE = None

    def __init__(
        self,
        name: Optional[str] = None,
        dryrun: bool = None,
    ):
        self._name, self._rns_folder = None, None
        if name is not None:
            # TODO validate that name complies with a simple regex
            if name.startswith("/builtins/"):
                name = name[10:]
            if name[0] == "^" and name != "^":
                name = name[1:]
            self._name, self._rns_folder = rns_client.split_rns_name_and_path(
                rns_client.resolve_rns_path(name)
            )

        self.dryrun = dryrun

    # TODO add a utility to allow a parameter to be specified as "default" and then use the default value

    @property
    def config_for_rns(self):
        config = {
            "resource_type": self.RESOURCE_TYPE,
            "resource_subtype": self.__class__.__name__,
        }
        config_attrs = ["name", "rns_address"]
        self.save_attrs_to_config(config, config_attrs)
        return config

    def _resource_string_for_subconfig(self, resource):
        """Returns a string representation of a sub-resource for use in a config."""
        if resource is None or isinstance(resource, str):
            return resource
        if resource.name:
            if resource.rns_address.startswith("^"):
                # Calls save internally and puts the resource in the current folder
                resource.name = rns_client.resolve_rns_path(resource.rns_address[1:])
            return resource.rns_address
        return resource.config_for_rns

    @property
    def rns_address(self):
        """Traverse up the filesystem until reaching one of the directories in rns_base_folders,
        then compute the relative path to that.

        Maybe later, account for folders along the path with a differnt RNS name."""

        if (
            self.name is None or self._rns_folder is None
        ):  # Anonymous folders have no rns address
            return None

        return str(Path(self._rns_folder) / self.name)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        # Split the name and rns path if path is given (concat with current_folder if just stem is given)
        self._name, self._rns_folder = split_rns_name_and_path(resolve_rns_path(name))

    @rns_address.setter
    def rns_address(self, new_address):
        self.name = new_address  # Note, this saves the resource to the new address!

    def save(
        self,
        name: str = None,
        snapshot: bool = False,
        overwrite: bool = True,
        **snapshot_kwargs,
    ):
        """Register the resource, saving it to local working_dir config and RNS config store. Uses the resource's
        `self.config_for_rns` to generate the dict to save."""

        # TODO deal with logic of saving anonymous folder for the first time after naming, i.e.
        # Path(tempfile.gettempdir()).relative_to(self.url) ...
        if name:
            if "/" in name[1:] or self._rns_folder is None:
                self._name, self._rns_folder = split_rns_name_and_path(
                    resolve_rns_path(name)
                )
            else:
                self._name = name

        # TODO handle self.access == 'read' instead of this weird overwrite argument
        save(self, snapshot=snapshot, overwrite=overwrite, **snapshot_kwargs)

        return self

    def __str__(self):
        return pprint.pformat(self.config_for_rns)

    @classmethod
    def from_name(cls, name, dryrun=False):
        config = rns_client.load_config(name=name)
        if not config:
            raise ValueError(f"Resource {name} not found.")
        config["name"] = name
        # Uses child class's from_config
        return cls.from_config(config=config, dryrun=dryrun)

    def unname(self):
        """Change the naming of the resource to anonymous and delete any local or RNS configs for the resource."""
        self.delete_configs()
        self._name = None

    @staticmethod
    def history(name: str, entries: int = 10) -> List[Dict]:
        """Return the history of this resource, including specific config fields (e.g. blob URL) and which runs
        have overwritten it."""
        resource_uri = rns_client.resource_uri(name)
        resp = requests.get(
            f"{rns_client.api_server_url}/resource/history/{resource_uri}",
            headers=rns_client.request_headers,
        )
        if resp.status_code != 200:
            raise Exception(
                f"Failed to load resource history: {json.loads(resp.content)}"
            )

        resource_history = read_response_data(resp)
        return resource_history

    # TODO delete sub-resources
    def delete_configs(self):
        """Delete the resource's config from local working_dir and RNS config store."""
        rns_client.delete_configs(resource=self)

    def save_attrs_to_config(self, config, attrs):
        for attr in attrs:
            val = self.__getattribute__(attr)
            if val:
                config[attr] = val

    # TODO [DG] Implement proper sharing of subresources (with an overload of some kind)
    def share(
        self, users: list, access_type: Union[ResourceAccess, str] = ResourceAccess.read
    ) -> Tuple[Dict[str, ResourceAccess], Dict[str, ResourceAccess]]:
        """Grant access to the resource for list of users. If a user has a Runhouse account they
        will receive an email notifying them of their new access. If the user does not have a Runhouse account they will
        also receive instructions on creating one, after which they will be able to have access to the Resource.
        Note: You can only grant resource access to other users if you have Write / Read privileges for the Resource.

        Args:
            users (list): list of user emails and / or runhouse account usernames.
            access_type (:obj:`ResourceAccess`, optional): access type to provide for the resource.

        Example:
            .. code-block:: python

               added_users, new_users = my_send.share(users=["username1", "user@gmail.com"], access_type='read')

        Returns:
            Tuple[Dict[str, ResourceAccess], Dict[str, ResourceAccess]]: Tuple of two dictionaries.

            `added_users`: users who already have an account and have been granted access to the resource.

            `new_users`: users who do not have Runhouse accounts.

        """
        if isinstance(access_type, str):
            access_type = ResourceAccess(access_type)

        if not rns_client.exists(self.rns_address):
            self.save()
        added_users, new_users = rns_client.grant_resource_access(
            resource_name=self.name, user_emails=users, access_type=access_type
        )
        return added_users, new_users



from typing import List, Optional, Union

from runhouse.rh_config import rns_client

from .cluster import Cluster
from .skycluster import SkyCluster


# Cluster factory method
def cluster(
    name: str,
    instance_type: Optional[str] = None,
    num_instances: Optional[int] = None,
    provider: Optional[str] = None,
    autostop_mins: Optional[int] = None,
    use_spot: Optional[bool] = None,
    image_id: Optional[str] = None,
    region: Optional[str] = None,
    ips: List[str] = None,
    ssh_creds: Optional[dict] = None,
    dryrun: Optional[bool] = False,
) -> Union[Cluster, SkyCluster]:
    config = rns_client.load_config(name)
    config["name"] = name or config.get("rns_address", None) or config.get("name")

    config["ips"] = ips or config.get("ips", None)
    # ssh creds should only be in Secrets management, not in config
    config["ssh_creds"] = ssh_creds or config.get("ssh_creds", None)
    if config["ips"]:
        return Cluster.from_config(config, dryrun=dryrun)

    config["instance_type"] = instance_type or config.get("instance_type", None)
    config["num_instances"] = num_instances or config.get("num_instances", None)
    config["provider"] = provider or config.get("provider", None)
    config["autostop_mins"] = (
        autostop_mins
        if autostop_mins is not None
        else config.get("autostop_mins", None)
    )
    config["use_spot"] = (
        use_spot if use_spot is not None else config.get("use_spot", None)
    )
    config["image_id"] = (
        image_id if image_id is not None else config.get("image_id", None)
    )
    config["region"] = region if region is not None else config.get("region", None)

    new_cluster = SkyCluster.from_config(config, dryrun=dryrun)

    return new_cluster



import copy
import logging
import uuid
from pathlib import Path
from typing import Dict, Optional

import runhouse as rh
from runhouse.rh_config import rns_client
from runhouse.rns.folders.folder import Folder, folder
from runhouse.rns.resource import Resource

logger = logging.getLogger(__name__)


class Blob(Resource):
    # TODO rename to "File" and take out serialization?
    RESOURCE_TYPE = "blob"
    DEFAULT_FOLDER_PATH = "/runhouse/blobs"

    def __init__(
        self,
        url: Optional[str] = None,
        name: Optional[str] = None,
        fs: Optional[str] = Folder.DEFAULT_FS,
        data_config: Optional[Dict] = None,
        dryrun: bool = False,
        **kwargs,
    ):
        """

        Args:
            name ():
            fs (): FSSpec protocol, e.g. 's3', 'gs'. See/run `fsspec.available_protocols()`.
                Default is "file", the local filesystem to wherever the blob is created.
            data_config ():
            serializer ():
        """
        super().__init__(name=name, dryrun=dryrun)
        self._filename = str(Path(url).name) if url else self.name
        # Use factory method so correct subclass for fs is returned
        self._folder = folder(
            url=str(Path(url).parent) if url is not None else url,
            fs=fs,
            data_config=data_config,
            dryrun=dryrun,
        )
        self._cached_data = None

    # TODO do we need a del?

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        blob_config = {
            "url": self.url,  # pair with data source to create the physical URL
            "resource_type": self.RESOURCE_TYPE,
            "fs": self.fs,
        }
        config.update(blob_config)
        return config

    @staticmethod
    def from_config(config: dict, dryrun=True):
        return Blob(**config, dryrun=dryrun)

    @property
    def data(self):
        """Get the blob data"""
        # TODO this caching is dumb, either get rid of it or replace with caching from fsspec
        if self._cached_data is not None:
            return self._cached_data
        data = self.fetch()
        return data

    @data.setter
    def data(self, new_data):
        """Update the data blob to new data"""
        self._cached_data = new_data
        # TODO should we save here?
        # self.save(overwrite=True)

    @property
    def fs(self):
        return self._folder.fs

    @fs.setter
    def fs(self, new_fs):
        self._folder.fs = new_fs

    @property
    def url(self):
        return self._folder.url + "/" + self._filename

    @url.setter
    def url(self, new_url):
        self._folder.url = str(Path(new_url).parent)
        self._filename = str(Path(new_url).name)

    @property
    def data_config(self):
        return self._folder.data_config

    @data_config.setter
    def data_config(self, new_data_config):
        self._folder.data_config = new_data_config

    @property
    def fsspec_url(self):
        return self._folder.fsspec_url + "/" + self._filename

    def open(self, mode="rb"):
        """Get a file-like (OpenFile container object) of the blob data. User must close the file, or use this
        method inside of a with statement (e.g. `with my_blob.open() as f:`)."""
        return self._folder.open(self._filename, mode=mode)

    def to(self, fs, url=None, data_config=None):
        new_table = copy.copy(self)
        new_table._folder = self._folder.to(fs=fs, url=url, data_config=data_config)
        return new_table

    def fetch(self):
        """Return the data for the user to deserialize"""
        self._cached_data = self._folder.get(self._filename)
        return self._cached_data

    def save(
        self,
        name: str = None,
        snapshot: bool = False,
        overwrite: bool = True,
        **snapshot_kwargs,
    ):

        # TODO figure out default behavior for not overwriting but still saving
        # if not overwrite:
        #     TODO check if data_url is already in use
        #     time = datetime.today().strftime('%Y-%m-%d_%H:%M:%S')
        #     self.data_url = self.data_url + time or time

        # TODO check if self._cached_data is None, and if so, don't just download it to then save it again?
        with self.open(mode="wb") as f:
            if not isinstance(self.data, bytes):
                # Avoid TypeError: a bytes-like object is required
                raise TypeError(
                    f"Cannot save blob with data of type {type(self.data)}, data must be serialized"
                )

            f.write(self.data)

        return super().save(
            name=name, snapshot=snapshot, overwrite=overwrite, **snapshot_kwargs
        )

    def delete_in_fs(self, recursive: bool = True):
        self._folder.rm(self._filename, recursive=recursive)

    def exists_in_fs(self):
        return self._folder.fsspec_fs.exists(self.fsspec_url)

    # TODO [DG] get rid of this in favor of just "sync_down(url, fs)" ?
    def sync_from_cluster(self, cluster, url: Optional[str] = None):
        """Efficiently rsync down a blob from a cluster, into the url of the current Blob object."""
        if not cluster.address:
            raise ValueError("Cluster must be started before copying data to it.")
        # TODO support fsspec urls (e.g. nonlocal fs's)?

        cluster.rsync(source=self.url, dest=url, up=False)

    def from_cluster(self, cluster):
        """Create a remote blob from a url on a cluster. This will create a virtual link into the
        cluster's filesystem. If you want to create a local copy or mount of the blob, use
        `Blob(url=<local_url>).sync_from_cluster(<cluster>, <url>)` or
        `Blob('url').from_cluster(<cluster>).mount(<local_url>)`."""
        if not cluster.address:
            raise ValueError("Cluster must be started before copying data from it.")
        new_blob = copy.deepcopy(self)
        new_blob._folder.fs = cluster
        return new_blob


def blob(
    data=None,
    name: Optional[str] = None,
    url: Optional[str] = None,
    fs: Optional[str] = None,
    data_config: Optional[Dict] = None,
    mkdir: bool = False,
    snapshot: bool = False,
    dryrun: bool = False,
):
    """Returns a Blob object, which can be used to interact with the resource at the given url

    Examples:
    # Creating the blob data - note the data should be provided as a serialized object, runhouse does not provide the
    # serialization functionality
    data = json.dumps(list(range(50))

    # 1. Create a remote blob with a name and no URL
    # provide a folder path for which to save in the remote file system
    # Since no URL is explicitly provided, we will save to a bucket called runhouse/blobs/my-blob
    rh.blob(name="@/my-blob", data=data, data_source='s3', dryrun=False)

    # 2. Create a remote blob with a name and URL
    rh.blob(name='@/my-blob', url='/runhouse-tests/my_blob.pickle', data=data, fs='s3', dryrun=False)

    # 3. Create a local blob with a name and a URL
    # save the blob to the local filesystem
    rh.blob(name=name, data=data, url=str(Path.cwd() / "my_blob.pickle"), dryrun=False)

    # 4. Create a local blob with a name and no URL
    # Since no URL is explicitly provided, we will save to ~/.cache/blobs/my-blob
    rh.blob(name="~/my-blob", data=data, dryrun=False)

    # Loading a blob
    my_local_blob = rh.blob(name="~/my_blob")
    my_s3_blob = rh.blob(name="@/my_blob")

    """
    config = rns_client.load_config(name)
    config["name"] = name or config.get("rns_address", None) or config.get("name")

    fs = fs or config.get("fs") or Folder.DEFAULT_FS
    config["fs"] = fs

    data_url = url or config.get("url")
    if data_url is None:
        # TODO [JL] move some of the default params in this factory method to the defaults module for configurability
        if fs == rns_client.DEFAULT_FS:
            # create random url to store in .cache folder of local filesystem
            data_url = str(Path(f"~/.cache/blobs/{uuid.uuid4().hex}").expanduser())
        else:
            # save to the default bucket
            name = name.lstrip(
                "/"
            )  # TODO [@JL] should we be setting config['name']=name again now?
            data_url = f"{Blob.DEFAULT_FOLDER_PATH}/{name}"

    config["url"] = data_url
    config["data_config"] = data_config or config.get("data_config")

    if mkdir:
        # create the remote folder for the blob
        folder_url = str(Path(data_url).parent)
        rh.folder(name=folder_url, fs=fs, dryrun=True).mkdir()

    new_blob = Blob.from_config(config, dryrun=dryrun)
    new_blob.data = data

    if new_blob.name and not dryrun:
        new_blob.save(snapshot=snapshot, overwrite=True)

    return new_blob



import contextlib
import logging
import subprocess
from pathlib import Path

import sky
import yaml
from sky.backends import backend_utils, CloudVmRayBackend

from runhouse.rh_config import configs, rns_client

from runhouse.rns.hardware.cluster import Cluster

logger = logging.getLogger(__name__)


class SkyCluster(Cluster):
    RESOURCE_TYPE = "cluster"

    def __init__(
        self,
        name,
        instance_type: str = None,
        num_instances: int = None,
        provider: str = None,
        dryrun=True,
        autostop_mins=None,
        use_spot=False,
        image_id=None,
        region=None,
        sky_data=None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Args:
            name:
            instance_type: Type of cloud instance to use for the cluster
            num_instances: Number of instances to use for the cluster
            provider: Cloud provider to use for the cluster
            dryrun:
            autostop_mins: Number of minutes to keep the cluster up for following inactivity, or
                -1 to keep up indefinitely.
        """

        super().__init__(name=name, dryrun=dryrun)

        self.instance_type = instance_type
        self.num_instances = num_instances
        self.provider = provider or configs.get("default_provider")
        self.autostop_mins = (
            autostop_mins
            if autostop_mins is not None
            else configs.get("default_autostop")
        )
        self.use_spot = use_spot if use_spot is not None else configs.get("use_spot")
        self.image_id = image_id
        self.region = region

        self.address = None
        self._yaml_path = None
        self._grpc_tunnel = None
        self._secrets_sent = False
        self.client = None
        self.sky_data = sky_data
        if self.sky_data is not None:
            self._save_sky_data()
        else:
            # Checks local SkyDB if cluster is up, and loads connection info if so.
            self.populate_vars_from_status(dryrun=True)

            # Cluster status is set to INIT in the Sky DB right after starting, so we need to refresh once
            if not self.address:
                self.populate_vars_from_status(dryrun=False)

    @staticmethod
    def from_config(config: dict, dryrun=False):
        return SkyCluster(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns

        # Also store the ssh keys for the cluster in RNS
        config.update(
            {
                "instance_type": self.instance_type,
                "num_instances": self.num_instances,
                "provider": self.provider,
                "autostop_mins": self.autostop_mins,
                "use_spot": self.use_spot,
                "image_id": self.image_id,
                "region": self.region,
                "sky_data": self._get_sky_data(),
            }
        )
        return config

    def _get_sky_data(self):
        config = sky.global_user_state.get_cluster_from_name(self.name)
        if not config:
            return None
        config["status"] = config[
            "status"
        ].name  # ClusterStatus enum is not json serializable
        if config["handle"]:
            with open(config["handle"].cluster_yaml, mode="r") as f:
                config["ray_config"] = yaml.safe_load(f)
            config["public_key"] = self.ssh_creds()["ssh_private_key"] + ".pub"
            config["handle"] = {
                "cluster_name": config["handle"].cluster_name,
                # This is saved as an absolute path - convert it to relative
                "cluster_yaml": self.relative_yaml_path(
                    yaml_path=config["handle"]._cluster_yaml
                ),
                "head_ip": config["handle"].head_ip,
                "launched_nodes": config["handle"].launched_nodes,
                "launched_resources": config[
                    "handle"
                ].launched_resources.to_yaml_config(),
            }
            config["handle"]["launched_resources"].pop("spot_recovery", None)

            pub_key_path = self.ssh_creds()["ssh_private_key"] + ".pub"
            if Path(pub_key_path).exists():
                with open(pub_key_path, mode="r") as f:
                    config["public_key"] = f.read()
        return config

    def _save_sky_data(self):
        # If we already have an entry for this cluster in the local sky files, ignore the new config
        # TODO [DG] when this is more stable maybe we shouldn't.
        yaml_path = self._yaml_path or self.sky_data.get("handle", {}).get(
            "cluster_yaml"
        )

        if (
            sky.global_user_state.get_cluster_from_name(self.name)
            and Path(yaml_path).expanduser().exists()
        ):
            return

        ray_config = self.sky_data.pop("ray_config", {})
        handle_info = self.sky_data.pop("handle", {})
        if not ray_config or not handle_info:
            raise Exception(
                "Expecting both `ray_config` and `handle` attributes in sky data"
            )

        if not Path(yaml_path).expanduser().parent.exists():
            Path(yaml_path).expanduser().parent.mkdir(parents=True, exist_ok=True)

        with Path(yaml_path).expanduser().open(mode="w+") as f:
            yaml.safe_dump(ray_config, f)

        cluster_abs_path = str(Path(yaml_path).expanduser())
        cloud_provider = sky.clouds.CLOUD_REGISTRY.from_str(
            handle_info["launched_resources"]["cloud"]
        )
        backend_utils._add_auth_to_cluster_config(cloud_provider, cluster_abs_path)

        resources = sky.Resources.from_yaml_config(handle_info["launched_resources"])
        handle = CloudVmRayBackend.ResourceHandle(
            cluster_name=self.name,
            cluster_yaml=cluster_abs_path,
            launched_nodes=handle_info["launched_nodes"],
            # head_ip=handle_info['head_ip'], # deprecated
            launched_resources=resources,
        )
        sky.global_user_state.add_or_update_cluster(
            cluster_name=self.name,
            cluster_handle=handle,
            requested_resources=[resources],
            is_launch=True,
            ready=False,
        )
        # Refresh the cluster status before saving the ssh info so SkyPilot has a chance to wipe the .ssh/config if
        # the cluster went down
        self.populate_vars_from_status(dryrun=False)
        if self.address:
            backend_utils.SSHConfigHelper.add_cluster(
                self.name, [handle_info["head_ip"]], ray_config["auth"]
            )

    def __getstate__(self):
        """Make sure sky_data is loaded in before pickling."""
        self.sky_data = self._get_sky_data()
        state = self.__dict__.copy()
        return state

    @staticmethod
    def relative_yaml_path(yaml_path):
        if Path(yaml_path).is_absolute():
            yaml_path = "~/.sky/generated/" + Path(yaml_path).name
        return yaml_path

    # ----------------- Launch/Lifecycle Methods -----------------

    # TODO [DG] this sometimes returns True when cluster is not up
    def is_up(self) -> bool:
        self.populate_vars_from_status(dryrun=False)
        return self.address is not None

    def status(self, refresh=True):
        """
        Return dict looks like:
         {'name': 'sky-cpunode-donny',
          'launched_at': 1662317201,
          'handle': ResourceHandle(
            cluster_name=sky-cpunode-donny,
            head_ip=54.211.97.164,
            cluster_yaml=/Users/donny/.sky/generated/sky-cpunode-donny.yml,
            launched_resources=1x AWS(m6i.2xlarge),
            tpu_create_script=None,
            tpu_delete_script=None),
          'last_use': 'sky cpunode',
          'status': <ClusterStatus.UP: 'UP'>,
          'autostop': -1,
          'metadata': {}}
        More: https://github.com/skypilot-org/skypilot/blob/0c2b291b03abe486b521b40a3069195e56b62324/sky/backends/cloud_vm_ray_backend.py#L1457
        """  # noqa
        # return backend_utils._refresh_cluster_record(
        #     self.name, force_refresh=refresh, acquire_per_cluster_status_lock=False
        # )
        return self.get_sky_statuses(cluster_name=self.name, refresh=refresh)

    def populate_vars_from_status(self, dryrun=False):
        # Try to get the cluster status from SkyDB
        cluster_dict = self.status(refresh=not dryrun)
        if not cluster_dict:
            return
        self.address = cluster_dict["handle"].head_ip
        self._yaml_path = cluster_dict["handle"].cluster_yaml
        if not cluster_dict["status"].name == "UP":
            self.address = None

    @staticmethod
    def get_sky_statuses(cluster_name: str = None, refresh: bool = True):
        """
        Get status dicts for all Sky clusters.
        Args:
            cluster_name (str): Return status dict for only specific cluster.

        Returns:

        """
        # TODO [DG] just get status for this cluster
        all_clusters_status = sky.status(refresh=refresh)
        if not cluster_name:
            return all_clusters_status
        for cluster_dict in all_clusters_status:
            if cluster_dict["name"] == cluster_name:
                return cluster_dict

    def up(
        self,
    ):
        if self.provider in ["aws", "gcp", "azure", "lambda", "cheapest"]:
            task = sky.Task(
                num_nodes=self.num_instances if ":" not in self.instance_type else None,
                # docker_image=image,  # Zongheng: this is experimental, don't use it
                # envs=None,
            )
            cloud_provider = (
                sky.clouds.CLOUD_REGISTRY.from_str(self.provider)
                if self.provider != "cheapest"
                else None
            )
            task.set_resources(
                sky.Resources(
                    cloud=cloud_provider,
                    instance_type=self.instance_type
                    if ":" not in self.instance_type and "CPU" not in self.instance_type
                    else None,
                    accelerators=self.instance_type
                    if ":" in self.instance_type and "CPU" not in self.instance_type
                    else None,
                    cpus=self.instance_type.rsplit(":", 1)[1]
                    if ":" in self.instance_type and "CPU" in self.instance_type
                    else None,
                    region=self.region,
                    image_id=self.image_id,
                    use_spot=self.use_spot,
                )
            )
            if Path("~/.rh").expanduser().exists():
                task.set_file_mounts(
                    {
                        "~/.rh": "~/.rh",
                    }
                )
            sky.launch(
                task,
                cluster_name=self.name,
                idle_minutes_to_autostop=self.autostop_mins,
                down=True,
            )
        elif self.provider == "k8s":
            # TODO ssh in and do for real
            # https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/config.html#kuberay-config
            # subprocess.Popen('kubectl apply -f raycluster.yaml'.split(' '))
            # self.address = cluster_dict['handle'].head_ip
            # self._yaml_path = cluster_dict['handle'].cluster_yaml
            raise NotImplementedError("Kubernetes Cluster provider not yet supported")
        else:
            raise ValueError(f"Cluster provider {self.provider} not supported.")

        self.populate_vars_from_status()
        self.restart_grpc_server()

    def keep_warm(self, autostop_mins=-1):
        sky.autostop(self.name, autostop_mins, down=True)
        self.autostop_mins = autostop_mins

    def teardown(self):
        # Stream logs
        sky.down(self.name)
        self.address = None

    def teardown_and_delete(self):
        self.teardown()
        rns_client.delete_configs()

    @contextlib.contextmanager
    def pause_autostop(self):
        sky.autostop(self.name, idle_minutes=-1)
        yield
        sky.autostop(self.name, idle_minutes=self.autostop_mins)

    # ----------------- SSH Methods ----------------- #

    @staticmethod
    def cluster_ssh_key(path_to_file):
        try:
            f = open(path_to_file, "r")
            private_key = f.read()
            return private_key
        except FileNotFoundError:
            raise Exception(f"File with ssh key not found in: {path_to_file}")

    def ssh_creds(self):
        if not self._yaml_path or not Path(self._yaml_path).exists():
            if self.sky_data:
                # If this cluster was serialized and sent over the wire, it will have sky_data (we make sure of that
                # in __getstate__) but no yaml, and we need to save down the sky data to the sky db and local yaml
                self._save_sky_data()
            else:
                # To avoid calling this twice (once in save_sky_data)
                self.populate_vars_from_status(dryrun=self.dryrun)

        return backend_utils.ssh_credential_from_yaml(self._yaml_path)

    def ssh(self):
        subprocess.run(["ssh", f"{self.name}"])



import copy
import logging
import os
import shutil
import subprocess
import tempfile
import uuid
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import fsspec

# TODO [DG] flip this when we switch to sshfs
import sshfs

import runhouse as rh
from runhouse.rh_config import configs, rns_client
from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.resource import Resource

fsspec.register_implementation("ssh", sshfs.SSHFileSystem)
# SSHFileSystem is not yet builtin.
# Line above suggested by fsspec devs: https://github.com/fsspec/filesystem_spec/issues/1071

logger = logging.getLogger(__name__)

PROVIDER_FS_LOOKUP = {
    "aws": "s3",
    "gcp": "gs",
    "azure": "abfs",
    "oracle": "ocifs",
    "databricks": "dbfs",
    "github": "github",
}


class Folder(Resource):
    RESOURCE_TYPE = "folder"
    DEFAULT_FS = "file"
    CLUSTER_FS = "ssh"
    DEFAULT_FOLDER_PATH = "/runhouse"
    DEFAULT_CACHE_FOLDER = "~/.cache/runhouse"

    def __init__(
        self,
        name: Optional[str] = None,
        url: Optional[str] = None,
        fs: Optional[str] = DEFAULT_FS,
        dryrun: bool = False,
        local_mount: bool = False,
        data_config: Optional[Dict] = None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        TODO [DG] Update
        Include loud warning that relative paths are relative to the git root / working directory!
        Args:
            name ():
            parent (): string path to parent folder, or
            data_source (): FSSpec protocol, e.g. 's3', 'gs'. See/run `fsspec.available_protocols()`.
                Default is "file", the local filesystem to wherever the blob is created.
            data_config ():
            local_path ():
        """
        super().__init__(name=name, dryrun=dryrun)

        self.fs = fs

        # TODO [DG] Should we ever be allowing this to be None?
        self._url = (
            self.default_url(self.rns_address, fs)
            if url is None
            else url
            if isinstance(fs, Resource)
            else url
            if Path(url).expanduser().is_absolute()
            else str(Path(rns_client.locate_working_dir()) / url)
        )
        self.data_config = data_config or {}

        self.local_mount = local_mount
        self._local_mount_path = None
        if local_mount:
            self.mount(tmp=True)
        if not self.dryrun:
            self.mkdir()

    @classmethod
    def default_url(cls, rns_address, fs):
        from runhouse.rns.hardware import Cluster

        if fs == Folder.DEFAULT_FS or isinstance(fs, Cluster):
            if rns_address:
                return str(
                    Path.cwd() / rns_client.split_rns_name_and_path(rns_address)[0]
                )  # saves to cwd / name
            return f"{Folder.DEFAULT_CACHE_FOLDER}/{uuid.uuid4().hex}"
        else:
            # If no URL provided for a remote file system default to its name if provided
            if rns_address:
                name = rns_address[1:].replace("/", "_") + f".{cls.RESOURCE_TYPE}"
                return f"{Folder.DEFAULT_FOLDER_PATH}/{name}"
            return f"{Folder.DEFAULT_FOLDER_PATH}/{uuid.uuid4().hex}"

    # ----------------------------------
    @staticmethod
    def from_config(config: dict, dryrun=True):
        """Load config values into the object."""
        if config["fs"] == "s3":
            from .s3_folder import S3Folder

            return S3Folder.from_config(config, dryrun=dryrun)
        elif config["fs"] == "gs":
            from .gcs_folder import GCSFolder

            return GCSFolder.from_config(config, dryrun=dryrun)
        elif config["fs"] == "azure":
            from .azure_folder import AzureFolder

            return AzureFolder.from_config(config, dryrun=dryrun)
        elif isinstance(config["fs"], dict):
            from runhouse.rns.hardware import Cluster

            config["fs"] = Cluster.from_config(config["fs"], dryrun=dryrun)
        return Folder(**config, dryrun=dryrun)

    @classmethod
    def from_name(cls, name, dryrun=False):
        config = rns_client.load_config(name=name)
        if not config:
            raise ValueError(f"Resource {name} not found.")

        config["name"] = name

        fs = config["fs"]
        if isinstance(fs, str) and fs.startswith("/"):
            # if the fs is set to a cluster
            cluster_config: dict = rns_client.load_config(name=fs)
            if not cluster_config:
                raise Exception(f"No cluster config saved for {fs}")

            # set the cluster config as the fs
            config["fs"] = cluster_config

        # Uses child class's from_config
        return cls.from_config(config=config, dryrun=dryrun)

    @property
    def url(self):
        if self._url is not None:
            if self.fs == Folder.DEFAULT_FS:
                return str(Path(self._url).expanduser())
            elif self._fs_str == self.CLUSTER_FS and self._url.startswith("~/"):
                # sftp takes relative urls to the home directory but doesn't understand '~'
                return self._url[2:]
            return self._url
        else:
            return None

    @url.setter
    def url(self, url):
        self._url = url
        self._local_mount_path = None

    @property
    def fs(self):
        return self._fs

    @fs.setter
    def fs(self, data_source):
        self._fs = data_source
        self._fsspec_fs = None

    # TODO figure out how to free sshfs properly (https://github.com/ronf/asyncssh/issues/112)
    # def __del__(self):
    #     if self.local_mount:
    #         self.unmount()
    #     if self._fsspec_fs and hasattr(self._fsspec_fs, "close"):
    #         self._fsspec_fs.close()

    @property
    def data_config(self):
        if isinstance(self.fs, Resource):  # if fs is a cluster
            if not self.fs.address:
                self.fs.populate_vars_from_status(dryrun=False)
                if not self.fs.address:
                    raise ValueError(
                        "Cluster must be started before copying data from it."
                    )
            creds = self.fs.ssh_creds()
            config_creds = {
                "host": self.fs.address,
                "username": creds["ssh_user"],
                # 'key_filename': str(Path(creds['ssh_private_key']).expanduser())}  # For SFTP
                "client_keys": [str(Path(creds["ssh_private_key"]).expanduser())],
            }  # For SSHFS
            ret_config = self._data_config.copy()
            ret_config.update(config_creds)
            return ret_config
        return self._data_config

    @data_config.setter
    def data_config(self, data_config):
        self._data_config = data_config
        self._fsspec_fs = None

    @property
    def _fs_str(self):
        if isinstance(self.fs, Resource):  # if fs is a cluster
            # TODO [DG] Return 'file' if we're on this cluster
            return self.CLUSTER_FS
        else:
            return self.fs

    @property
    def fsspec_fs(self):
        if self._fsspec_fs is None:
            self._fsspec_fs = fsspec.filesystem(self._fs_str, **self.data_config)
        return self._fsspec_fs

    @property
    def local_path(self):
        if self.is_local():
            return self._local_mount_path or str(Path(self.url).expanduser())
        else:
            return None

    def is_writable(self):
        # If the filesystem hasn't overridden mkdirs, it's a no-op and the filesystem is probably readonly
        # (e.g. https://filesystem-spec.readthedocs.io/en/latest/_modules/fsspec/implementations/github.html).
        # In that case, we should just create a new folder in the default
        # location and add it as a child to the parent folder.
        return self.fsspec_fs.__class__.mkdirs == fsspec.AbstractFileSystem.mkdirs

    def mv(self, fs, url=None, data_config=None) -> None:
        """Move the folder to a new filesystem.

        Args:
            fs (str): file system.
            url (:obj:`str`, optional): fsspec URL.
            data_config(:obj:`dict`, optional): Config to move.
        """
        # TODO [DG] create get_default_url for fs method to be shared
        if url is None:
            url = "rh/" + self.rns_address
        data_config = data_config or {}
        with fsspec.open(self.fsspec_url, **self.data_config) as src:
            with fsspec.open(f"{fs}://{url}", **data_config) as dest:
                # NOTE For packages, maybe use the `ignore` param here to only copy python files.
                new_url = shutil.move(src, dest)
        self.url = new_url
        self.fs = fs
        self.data_config = data_config or {}

    def to(self, fs, url: Optional[str] = None, data_config: Optional[dict] = None):
        """Copy the folder to a new filesystem, and return a new Folder object pointing to the new location."""
        # silly syntactic sugar to allow `my_remote_folder.to('here')`, clearer than `to('file')`
        if fs == "here":
            fs = "file"
            url = str(Path.cwd() / self.url.split("/")[-1]) if url is None else url

        url = str(
            url or self.default_url(self.name, fs)
        )  # Make sure it's a string and not a Path

        fs_str = getattr(
            fs, "name", fs
        )  # Use fs.name if available, i.e. fs is a cluster
        logging.info(
            f"Copying folder from {self.fsspec_url} to: {fs_str}, with url: {url}"
        )

        # to_local, to_cluster and to_data_store are also overridden by subclasses to dispatch
        # to more performant cloud-specific APIs
        from runhouse.rns.hardware import Cluster

        if fs == "file":
            return self.to_local(
                dest_url=url, data_config=data_config, return_dest_folder=True
            )
        elif isinstance(fs, Cluster):  # If fs is a cluster
            # TODO [DG] change default behavior to return_dest_folder=False
            return self.to_cluster(dest_cluster=fs, url=url, return_dest_folder=True)
        elif fs in ["s3", "gs", "azure"]:
            return self.to_data_store(
                fs=fs, data_store_url=url, data_config=data_config
            )
        else:
            self.fsspec_copy(fs, url, data_config)
            new_folder = copy.deepcopy(self)
            new_folder.url = url
            new_folder.fs = fs
            new_folder.data_config = data_config or {}
            return new_folder

    def fsspec_copy(self, fs: str, url: str, data_config: dict):
        # Fallback for other fsspec filesystems, but very slow:
        if self.is_local():
            self.fsspec_fs.put(self.url, f"{fs}://{url}", recursive=True)
        else:
            # TODO this is really really slow, maybe use skyplane, as follows:
            # src_url = f'local://{self.url}' if self.is_local() else self.fsspec_url
            # subprocess.run(['skyplane', 'sync', src_url, f'{fs}://{url}'])

            # FYI: from https://github.com/fsspec/filesystem_spec/issues/909
            # TODO [DG]: Copy chunks https://github.com/fsspec/filesystem_spec/issues/909#issuecomment-1204212507
            src = fsspec.get_mapper(self.fsspec_url, create=False, **self.data_config)
            dest = fsspec.get_mapper(f"{fs}://{url}", create=True, **data_config)
            # dest.fs.mkdir(dest.root, create_parents=True)
            import tqdm

            for k in tqdm.tqdm(src):
                # NOTE For packages, maybe use the `ignore` param here to only copy python files.
                dest[k] = src[k]
                # dst.write(src.read())

    def destination_folder(
        self,
        dest_url: str,
        dest_fs: Optional[str] = "file",
        data_config: Optional[dict] = None,
    ):
        """Return a new Folder object pointing to the destination folder."""
        new_folder = copy.deepcopy(self)
        new_folder.url = dest_url
        new_folder.fs = dest_fs
        new_folder.data_config = data_config or {}
        return new_folder

    def to_local(
        self, dest_url: str, data_config: dict, return_dest_folder: bool = False
    ):
        from runhouse.rns.hardware import Cluster

        if self.fs == "file":
            # Simply move the files within local fs
            shutil.copytree(src=self.url, dst=dest_url)
        elif isinstance(self.fs, Cluster):
            return self.from_cluster(cluster=self.fs, dest_url=dest_url)
        else:
            self.fsspec_copy("file", dest_url, data_config)

        if return_dest_folder:
            return self.destination_folder(
                dest_url=dest_url, dest_fs="file", data_config=data_config
            )

    # TODO [DG] Any reason to keep this?
    # def to_sftp(self, url, data_config):
    #     from runhouse.rns.hardware import Cluster
    #     if self.fs == 'file':
    #         # Rsync up the files to the remote fs
    #         self.rsync(local=self.url, remote=url, data_config=data_config, up=True)
    #     elif self.fs == 'sftp':
    #         # Simply move the files within stfp fs
    #         # TODO [DG] speculation
    #         self.fsspec_fs.mv(self.url, url)
    #     elif isinstance(self.fs, Cluster):
    #         self.fs.run([f'rsync {self.url} {data_config["username"]}@{data_config["host"]}:{url} '
    #                      f'--password_file {data_config["key_filename"]}'])
    #     else:
    #         self.fsspec_copy('file', url, data_config)

    def to_data_store(
        self,
        fs: str,
        data_store_url: Optional[str] = None,
        data_config: Optional[dict] = None,
        return_dest_folder: bool = True,
    ):
        """Local or cluster to blob storage"""
        from runhouse.rns.hardware import Cluster

        local_folder_url = self.url

        folder_config = self.config_for_rns
        folder_config["fs"] = fs
        folder_config["url"] = data_store_url
        folder_config["data_config"] = data_config
        new_folder = Folder.from_config(folder_config)

        if self.fs == "file":
            new_folder.upload(src=local_folder_url)
        elif isinstance(self.fs, Cluster):
            self.fs.run(
                [new_folder.upload_command(src=local_folder_url, dest=new_folder.url)]
            )
        else:
            self.fsspec_copy("file", data_store_url, data_config)

        return new_folder

    @staticmethod
    def rsync(local, remote, data_config, up=True):
        # TODO convert this to generate rsync command between two clusters
        dest_str = f'{data_config["username"]}@{data_config["host"]}:{remote}'
        src_str = local
        if not up:
            src_str, dest_str = dest_str, src_str
        subprocess.check_call(
            f"rsync {src_str} {dest_str} "
            f'--password_file {data_config["key_filename"]}'
        )

    def mkdir(self):
        """create the folder in specified file system if it doesn't already exist"""
        folder_url = self.url
        if Path(os.path.basename(folder_url)).suffix != "":
            folder_url = str(Path(folder_url).parent)

        logging.info(f"Creating new {self._fs_str} folder: {folder_url}")
        self.fsspec_fs.mkdirs(folder_url, exist_ok=True)

    def mount(self, url: Optional[str] = None, tmp: bool = False) -> str:
        """Mount the folder locally."""
        # TODO check that fusepy and FUSE are installed
        if tmp:
            self._local_mount_path = tempfile.mkdtemp()
        else:
            self._local_mount_path = url
        remote_fs = self.fsspec_fs
        fsspec.fuse.run(fs=remote_fs, path=self.url, mount_point=self._local_mount_path)
        return self._local_mount_path

    def to_cluster(self, dest_cluster, url=None, mount=False, return_dest_folder=True):
        """Copy the folder from a file or cluster source onto a cluster."""
        if not dest_cluster.address:
            raise ValueError("Cluster must be started before copying data to it.")

        # Create tmp_mount if needed
        if not self.is_local() and mount:
            self.mount(tmp=True)

        dest_url = url or f"~/{Path(self.url).stem}"

        # Need to add slash for rsync to copy the contents of the folder
        dest_folder = copy.deepcopy(self)
        dest_folder.url = dest_url
        dest_folder.fs = dest_cluster
        dest_folder.mkdir()

        if self.fs == "file":
            dest_cluster.rsync(source=self.url, dest=dest_url, up=True, contents=True)

        elif isinstance(self.fs, Resource):
            src_url = self.url

            cluster_creds = self.fs.ssh_creds()
            creds_file = cluster_creds["ssh_private_key"]

            command = (
                f"rsync -Pavz --filter='dir-merge,- .gitignore' -e \"ssh -i '{creds_file}' "
                f"-o StrictHostKeyChecking=no -o IdentitiesOnly=yes -o ExitOnForwardFailure=yes "
                f"-o ServerAliveInterval=5 -o ServerAliveCountMax=3 -o ConnectTimeout=30s -o ForwardAgent=yes "
                f'-o ControlMaster=auto -o ControlPersist=300s" {src_url}/ {dest_cluster.address}:{dest_url}'
            )
            status_codes = self.fs.run([command])
            if status_codes[0][0] != 0:
                raise Exception(
                    f"Error syncing folder to destination cluster ({dest_cluster.name}). "
                    f"Make sure the source cluster ({self.fs.name}) has the sky keys "
                    f"loaded in path: {creds_file}. "
                    f"For example: `my_cluster.send_secrets(providers=['sky'])`"
                )

        else:
            raise TypeError(
                f"`to_cluster` not supported for filesystem type {type(self.fs)}"
            )

        if return_dest_folder:
            dest_folder.fs = "file"

        return dest_folder

    def from_cluster(self, cluster, dest_url=None):
        """Create a remote folder from a url on a cluster.

        If `dest_url=None`, this will not perform any copy, and simply convert the resource to have a remote
        sftp filesystem into the cluster. If `dest_url` is set, it will rsync down the data and return a folder
        with fs=='file'.

        """
        if dest_url:
            if not cluster.address:
                raise ValueError("Cluster must be started before copying data from it.")
            # TODO support fsspec urls (e.g. nonlocal fs's)?
            Path(dest_url).expanduser().mkdir(parents=True, exist_ok=True)
            cluster.rsync(
                source=self.url,
                dest=str(Path(dest_url).expanduser()),
                up=False,
                contents=True,
            )
            new_folder = copy.deepcopy(self)
            new_folder.url = dest_url
            new_folder.fs = "file"
            # Don't need to do anything with _data_config because cluster creds are injected virtually through the
            # data_config property
            return new_folder
        else:
            new_folder = copy.deepcopy(self)
            new_folder.fs = cluster
            return new_folder

    def is_local(self):
        return (
            self.fs == "file" and self.url is not None and Path(self.url).exists()
        ) or self._local_mount_path

    def share(
        self,
        users: list,
        access_type: Union[ResourceAccess, str] = ResourceAccess.read,
        snapshot: bool = True,
        snapshot_fs: str = None,
        snapshot_compression: str = None,
        snapshot_url: str = None,
    ) -> Tuple[Dict[str, ResourceAccess], Dict[str, ResourceAccess]]:
        """Granting access to the resource for list of users (via their emails). If a user has a Runhouse account they
        will receive an email notifying them of their new access. If the user does not have a Runhouse account they will
        also receive instructions on creating one, after which they will be able to have access to the Resource.
        Note: You can only grant resource access to other users if you have Write / Read privileges for the Resource"""
        if self.is_local() and snapshot:
            # raise ValueError('Cannot share a local resource.')
            fs = snapshot_fs or PROVIDER_FS_LOOKUP[configs.get("default_provider")]
            if fs not in fsspec.available_protocols():
                raise ValueError(
                    f"Invalid mount_fs: {snapshot_fs}. Must be one of {fsspec.available_protocols()}"
                )
            if snapshot_compression not in fsspec.available_compressions():
                raise ValueError(
                    f"Invalid mount_compression: {snapshot_compression}. Must be one of "
                    f"{fsspec.available_compressions()}"
                )
            data_config = (
                {"compression": snapshot_compression} if snapshot_compression else {}
            )
            snapshot_folder = self.to(fs=fs, url=snapshot_url, data_config=data_config)

            # Is this a bad idea? Better to store the snapshot config as the source of truth than the local url
            rns_address = rns_client.local_to_remote_address(self.rns_address)
            snapshot_folder.save(name=rns_address)

            return snapshot_folder.share(
                users=users, access_type=access_type, snapshot=False
            )

        # TODO just call super().share
        if isinstance(access_type, str):
            access_type = ResourceAccess(access_type)
        if not rns_client.exists(self.rns_address):
            self.save(name=rns_client.local_to_remote_address(self.rns_address))
        added_users, new_users = rns_client.grant_resource_access(
            resource_name=self.name, user_emails=users, access_type=access_type
        )
        return added_users, new_users

    def empty_folder(self):
        """Remove folder contents, but not the folder itself."""
        for p in self.fsspec_fs.ls(self.url):
            self.fsspec_fs.rm(p)

    def upload(self, src: str, region: Optional[str] = None):
        """Upload a folder to a remote bucket."""
        raise NotImplementedError

    def upload_command(self, src: str, dest: str):
        """CLI command for uploading folder to remote bucket. Needed when uploading a folder from a cluster."""
        raise NotImplementedError

    def run_upload_cli_cmd(self, sync_dir_command: str, access_denied_message: str):
        """Uploads a folder to a remote bucket.
        Based on the CLI command skypilot uses to upload the folder"""
        from sky.data.data_utils import run_upload_cli

        run_upload_cli(
            command=sync_dir_command,
            access_denied_message=access_denied_message,
            bucket_name=self.bucket_name_from_url(self.url),
        )

    def download(self, dest):
        raise NotImplementedError

    def download_command(self, src, dest):
        """CLI command for downloading folder from remote bucket. Needed when downloading a folder to a cluster."""
        raise NotImplementedError

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        config_attrs = ["local_mount", "data_config"]
        self.save_attrs_to_config(config, config_attrs)

        if self.fs == Folder.DEFAULT_FS:
            # If folder is local check whether path is relative, and if so take it relative to the working director
            # rather than to the home directory. If absolute, it's left alone.
            config["url"] = (
                self._url_relative_to_rh_workdir(self.url) if self.url else None
            )
        else:
            # if not a local filesystem save path as is (e.g. bucket/path)
            config["url"] = self.url

        if isinstance(self.fs, Resource):  # If fs is a cluster
            config["fs"] = self._resource_string_for_subconfig(self.fs)
        else:
            config["fs"] = self.fs

        return config

    @staticmethod
    def _url_relative_to_rh_workdir(url):
        rh_workdir = Path(rns_client.locate_working_dir())
        try:
            return str(Path(url).relative_to(rh_workdir))
        except ValueError:
            return url

    @property
    def fsspec_url(self):
        """Generate the FSSpec URL using the file system and url of the folder"""
        if self.url.startswith("/") and self._fs_str not in [
            rns_client.DEFAULT_FS,
            self.CLUSTER_FS,
        ]:
            return f"{self._fs_str}:/{self.url}"
        else:
            # For local, ssh / sftp filesystems we need both slashes
            # e.g.: 'ssh:///home/ubuntu/.cache/runhouse/tables/dede71ef83ce45ffa8cb27d746f97ee8'
            return f"{self._fs_str}://{self.url}"

    def ls(self, full_paths: bool = True):
        """List the contents of the folder"""
        paths = self.fsspec_fs.ls(path=self.url) if self.url else []
        if full_paths:
            return paths
        else:
            return [Path(path).name for path in paths]

    def resources(self, full_paths: bool = False, resource_type: str = None):
        """List the resources in the *RNS* folder.

        Args:
            full_paths (bool): If True, return the full RNS path for each resource. If False, return the
                resource name only.
            resource_type (str): If provided, only return resources of the specified type.
        """
        # TODO filter by type
        # TODO allow '*' wildcard for listing all resources (and maybe other wildcard things)
        try:
            resources = [
                path for path in self.ls() if (Path(path) / "config.json").exists()
            ]
        except FileNotFoundError:
            return []

        if full_paths:
            return [self.rns_address + "/" + Path(path).stem for path in resources]
        else:
            return [Path(path).stem for path in resources]

    @property
    def rns_address(self):
        """Traverse up the filesystem until reaching one of the directories in rns_base_folders,
        then compute the relative path to that.

        Maybe later, account for folders along the path with a different RNS name."""

        if self.name is None:  # Anonymous folders have no rns address
            return None

        # Only should be necessary when a new base folder is being added (therefore isn't in rns_base_folders yet)
        if self._rns_folder:
            return str(Path(self._rns_folder) / self.name)

        if self.url in rns_client.rns_base_folders.values():
            if self._rns_folder:
                return self._rns_folder + "/" + self.name
            else:
                return rns_client.default_folder + "/" + self.name

        segment = Path(self.url)
        while (
            str(segment) not in rns_client.rns_base_folders.values()
            and not segment == Path.home()
            and not segment == segment.parent
        ):
            segment = segment.parent

        if (
            segment == Path.home() or segment == segment.parent
        ):  # TODO throw an error instead?
            return rns_client.default_folder + "/" + self.name
        else:
            base_folder = Folder(url=str(segment), dryrun=True)
            base_folder_path = base_folder.rns_address
            relative_path = str(Path(self.url).relative_to(base_folder.url))
            return base_folder_path + "/" + relative_path

    def contains(self, name_or_path) -> bool:
        url, fs = self.locate(name_or_path)
        return url is not None

    def locate(self, name_or_path) -> (str, str):
        """Locate the local url of a Folder given an rns path.

        Keep in mind we're using both _rns_ path and physical path logic below. Be careful!
        """

        # If the path is already given relative to the current folder:
        if (Path(self.url) / name_or_path).exists():
            return str(Path(self.url) / name_or_path), self.fs

        # If name or path uses ~/ or ./, need to resolve with folder url
        abs_path = rns_client.resolve_rns_path(name_or_path)
        rns_path = self.rns_address

        # If this folder is anonymous, it has no rns contents
        if rns_path is None:
            return None, None

        if abs_path == rns_path:
            return self.url, self.fs
        try:
            child_url = Path(self.url) / Path(abs_path).relative_to(rns_path)
            if child_url.exists():
                return str(child_url), self.fs
        except ValueError:
            pass

        # Last resort, recursively search inside sub-folders and children.

        segments = abs_path.lstrip("/").split("/")
        if len(segments) == 1:
            return (
                None,
                None,
            )  # If only a single element, would have been found in ls above.

        # Look for lowest folder in the path that exists in filesystem, and recurse from that folder
        greatest_common_folder = Path(self.url)
        i = 0
        for i, seg in enumerate(segments):
            if not (greatest_common_folder / seg).exists():
                break
            greatest_common_folder = greatest_common_folder / seg
        if not str(greatest_common_folder) == self.url:
            return Folder(
                url=str(greatest_common_folder), fs=self.fs, dryrun=True
            ).locate("/".join(segments[i + 1 :]))

        return None, None

    def open(self, name, mode="rb", encoding=None):
        """Returns an fsspec file, which must be used as a content manager to be opened!
        e.g. with my_folder.open('obj_name') as my_file:
                pickle.load(my_file)
        """
        return self.fsspec_fs.open(self.url + "/" + name, mode=mode, encoding=encoding)

    def get(self, name, mode="rb", encoding=None):
        """Returns the contents of a file as a string or bytes."""
        with self.open(name, mode=mode, encoding=encoding) as f:
            return f.read()

    # TODO [DG] fix this to follow the correct convention above
    def get_all(self):
        # TODO we're not closing these, do we need to extract file-like objects so we can close them?
        return fsspec.open_files(self.fsspec_url, mode="rb", **self.data_config)

    def exists_in_fs(self):
        return self.fsspec_fs.exists(
            self.fsspec_url
        ) or rh.rns.top_level_rns_fns.exists(self.url)

    def delete_in_fs(self, recursive: bool = True):
        try:
            self.fsspec_fs.rmdir(self.url)
        except Exception as e:
            raise Exception(f"Failed to delete from file system: {e}")

    def rm(self, name, recursive: bool = True):
        """Remove a resource from the folder."""
        try:
            self.fsspec_fs.rm(self.fsspec_url + "/" + name, recursive=recursive)
        except FileNotFoundError:
            pass

    def put(self, contents, overwrite=False):
        """
        files: Either 1) A dict with keys being the file names and values being the
            file-like objects to write, 2) a Resource, or 3) a list of Resources.
        """
        # TODO create the bucket if it doesn't already exist
        # Handle lists of resources just for convenience
        if isinstance(contents, list):
            for resource in contents:
                self.put(resource, overwrite=overwrite)
            return

        if isinstance(contents, Folder):
            if not self.is_writable():
                raise RuntimeError(
                    f"Cannot put files into non-writable folder {self.name or self.url}"
                )
            if contents.url is None:  # Should only be the case when Folder is created
                contents.url = self.url + "/" + contents.name
                contents.fs = self.fs
                # The parent can be anonymous, e.g. the 'rh' folder.
                # TODO not sure if this should be allowed - if parent folder has no rns address, why would child
                # just be put into the default rns folder?
                # TODO If the base is named later, figure out what to do with the contents (rename, resave, etc.).
                if self.rns_address is None:
                    contents.rns_path = rns_client.default_folder + "/" + contents.name
                    rns_client.rns_base_folders.update(
                        {contents.rns_address: contents.url}
                    )
                # We don't need to call .save here to write down because it will be called at the end of the
                # folder or resource constructor
            else:
                if contents.name is None:  # Anonymous resource
                    i = 1
                    new_name = contents.RESOURCE_TYPE + str(i)
                    # Resolve naming conflicts if necessary
                    while rns_client.exists(self.url + "/" + new_name):
                        i += 1
                        new_name = contents.RESOURCE_TYPE + str(i)
                else:
                    new_name = contents.name

                # NOTE For intercloud transfer, we should use Skyplane
                with fsspec.open(
                    self.fsspec_url + "/" + new_name, **self.data_config
                ) as dest:
                    with fsspec.open(
                        contents.fsspec_url, **contents.data_config
                    ) as src:
                        # NOTE For packages, maybe use the `ignore` param here to only copy python files.
                        shutil.move(src, dest)
            return

        if not isinstance(contents, dict):
            raise TypeError(
                "`files` argument to `.put` must be Resource, list of Resources, or dict mapping "
                "filenames to file-like-objects"
            )

        if overwrite is False:
            folder_contents = self.resources()
            intersection = set(folder_contents).intersection(set(contents.keys()))
            if intersection != set():
                raise FileExistsError(
                    f"File(s) {intersection} already exist(s) at url"
                    f"{self.url}, cannot save them without overwriting."
                )
        # TODO figure out default behavior for not overwriting but still saving
        # if not overwrite:
        #     time = datetime.today().strftime('%Y-%m-%d_%H:%M:%S')
        #     self.data_url = self.data_url + time or time
        filenames = list(contents)
        fss_files = fsspec.open_files(
            self.fsspec_url + "/*",
            mode="wb",
            **self.data_config,
            num=len(contents),
            name_function=filenames.__getitem__,
        )
        for (fss_file, raw_file) in zip(fss_files, contents.values()):
            with fss_file as f:
                f.write(raw_file)

    @staticmethod
    def bucket_name_from_url(url: str) -> str:
        """Extract the bucket name from a URL (e.g. '/my-bucket/my-folder/my-file.txt' -> 'my-bucket')"""
        return Path(url).parts[1]


def folder(
    name: Optional[str] = None,
    url: Optional[Union[str, Path]] = None,
    fs: Optional[str] = None,
    dryrun: bool = False,
    local_mount: bool = False,
    data_config: Optional[Dict] = None,
):
    """Returns a folder object, which can be used to interact with the folder at the given url.
    The folder will be saved if `dryrun` is False.
    """
    config = rns_client.load_config(name)
    config["name"] = name or config.get("rns_address", None) or config.get("name")
    config["url"] = url or config.get("url")
    config["local_mount"] = local_mount or config.get("local_mount")
    config["data_config"] = data_config or config.get("data_config")

    file_system = fs or config.get("fs") or Folder.DEFAULT_FS
    config["fs"] = file_system
    if isinstance(file_system, str):
        if file_system in ["file", "github", "sftp", "ssh"]:
            new_folder = Folder.from_config(config, dryrun=dryrun)
        elif file_system == "s3":
            from .s3_folder import S3Folder

            new_folder = S3Folder.from_config(config, dryrun=dryrun)
        elif file_system == "gs":
            from .gcs_folder import GCSFolder

            new_folder = GCSFolder.from_config(config, dryrun=dryrun)
        elif file_system == "azure":
            from .azure_folder import AzureFolder

            new_folder = AzureFolder.from_config(config, dryrun=dryrun)
        elif file_system in fsspec.available_protocols():
            logger.warning(
                f"fsspec file system {file_system} not officially supported. Use at your own risk."
            )
            new_folder = Folder.from_config(config, dryrun=dryrun)
        elif rns_client.exists(file_system, resource_type="cluster"):
            config["fs"] = rns_client.load_config(file_system)
        else:
            raise ValueError(
                f"File system {file_system} not found. Have you installed the "
                f"necessary packages for this fsspec protocol? (e.g. s3fs for s3). If the file system "
                f"is a cluster (ex: /my-user/rh-cpu), make sure the cluster config has been saved."
            )

    # If cluster is passed as the fs.
    if isinstance(config["fs"], dict) or isinstance(
        config["fs"], Resource
    ):  # if fs is a cluster
        new_folder = Folder.from_config(config, dryrun=dryrun)

    return new_folder
